@@ -1,6 +1,6 @@
 /*
  * Copyright (c) [2011-2014] Novell, Inc.
- * Copyright (c) [2016,2018] SUSE LLC
+ * Copyright (c) [2016-2020] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -31,6 +31,7 @@
 
 #include "utils/Range.h"
 #include "utils/equal-date.h"
+#include "utils/HumanString.h"
 #include "cleanup.h"
 
 
@@ -46,6 +47,7 @@ struct Parameters
 
     time_t min_age;
     double space_limit;
+    double free_limit;
 
     template<typename Type>
     void read(const ProxyConfig& config, const char* name, Type& value)
@@ -59,19 +61,21 @@ struct Parameters
 
 
 Parameters::Parameters(const ProxySnapper* snapper)
-    : min_age(1800), space_limit(0.5)
+    : min_age(1800), space_limit(0.5), free_limit(0.2)
 {
     ProxyConfig config = snapper->getConfig();
 
     read(config, "SPACE_LIMIT", space_limit);
+    read(config, "FREE_LIMIT", free_limit);
 }
 
 
 ostream&
 operator<<(ostream& s, const Parameters& parameters)
 {
-    return s << "min-age:" << parameters.min_age << endl
-	     << "space-limit:" << parameters.space_limit;
+    return s << "min-age:" << parameters.min_age << '\n'
+	     << "space-limit:" << parameters.space_limit << '\n'
+	     << "free-limit:" << parameters.free_limit;
 }
 
 
@@ -85,11 +89,12 @@ public:
     virtual ~Cleaner() {}
 
     void cleanup();
+    void cleanup(std::function<bool()> condition);
 
 protected:
 
     virtual list<ProxySnapshots::iterator> calculate_candidates(ProxySnapshots& snapshots,
-								Range::Value value) = 0;
+								const Range::Value& value) = 0;
 
     struct younger_than
     {
@@ -114,15 +119,24 @@ protected:
 
     void remove(const list<ProxySnapshots::iterator>& tmp);
 
+    // Should the cleanup with quota space be run?
     bool is_quota_aware() const;
+
+    // Is the quota space condition satisfied?
     bool is_quota_satisfied() const;
 
-    void cleanup_quota_unaware(ProxySnapshots& snapshots);
-    void cleanup_quota_aware(ProxySnapshots& snapshots);
+    // Should the cleanup with free space be run?
+    bool is_free_aware() const;
+
+    // Is the free space condition satisfied?
+    bool is_free_satisfied() const;
+
+    void cleanup(ProxySnapshots& snapshots);
+    void cleanup(ProxySnapshots& snapshots, std::function<bool()> condition);
 
     ProxySnapper* snapper;
-    bool verbose;
 
+    const bool verbose;
     const Parameters& parameters;
 
 };
@@ -223,18 +237,6 @@ Cleaner::is_quota_aware() const
     {
 	snapper->prepareQuota();
     }
-    catch (const DBus::ErrorException& e)
-    {
-	SN_CAUGHT(e);
-
-	if (strcmp(e.name(), "error.quota") == 0)
-	{
-	    cerr << "quota not working (" << e.message() << ")" << endl;
-	    return false;
-	}
-
-	SN_RETHROW(e);
-    }
     catch (const QuotaException& e)
     {
 	SN_CAUGHT(e);
@@ -243,7 +245,7 @@ Cleaner::is_quota_aware() const
 	return false;
     }
 
-    return true;
+    return parameters.space_limit < 1.0;
 }
 
 
@@ -252,12 +254,69 @@ Cleaner::is_quota_satisfied() const
 {
     QuotaData quota_data = snapper->queryQuotaData();
 
-    return quota_data.used < parameters.space_limit * quota_data.size;
+    if (quota_data.size == 0)
+	return true;
+
+    double fraction = (double)(quota_data.used) / (double)(quota_data.size);
+
+    bool satisfied = fraction < parameters.space_limit;
+
+#ifdef VERBOSE_LOGGING
+    cout << byte_to_humanstring(quota_data.size, 2) << ", "
+	 << byte_to_humanstring(quota_data.used, 2) << ", "
+	 << fraction << ", " << satisfied << '\n';
+#endif
+
+    return satisfied;
+}
+
+
+bool
+Cleaner::is_free_aware() const
+{
+    if (parameters.is_degenerated())
+	return false;
+
+    try
+    {
+	snapper->queryFreeSpaceData();
+    }
+    catch (const FreeSpaceException& e)
+    {
+	SN_CAUGHT(e);
+
+	cerr << "free space not working (" << e.what() << ")" << endl;
+	return false;
+    }
+
+    return parameters.free_limit > 0.0;
+}
+
+
+bool
+Cleaner::is_free_satisfied() const
+{
+    FreeSpaceData free_space_data = snapper->queryFreeSpaceData();
+
+    if (free_space_data.size == 0)
+	return true;
+
+    double fraction = (double)(free_space_data.free) / (double)(free_space_data.size);
+
+    bool satisfied = fraction > parameters.free_limit;
+
+#ifdef VERBOSE_LOGGING
+    cout << byte_to_humanstring(free_space_data.size, 2) << ", "
+	 << byte_to_humanstring(free_space_data.free, 2) << ", "
+	 << fraction << ", " << satisfied << '\n';
+#endif
+
+    return satisfied;
 }
 
 
 void
-Cleaner::cleanup_quota_unaware(ProxySnapshots& snapshots)
+Cleaner::cleanup(ProxySnapshots& snapshots)
 {
     list<ProxySnapshots::iterator> candidates = calculate_candidates(snapshots, Range::MAX);
 
@@ -268,14 +327,19 @@ Cleaner::cleanup_quota_unaware(ProxySnapshots& snapshots)
 
 
 void
-Cleaner::cleanup_quota_aware(ProxySnapshots& snapshots)
+Cleaner::cleanup(ProxySnapshots& snapshots, std::function<bool()> condition)
 {
-    while (!is_quota_satisfied())
+    while (!condition())
     {
 	list<ProxySnapshots::iterator> candidates = calculate_candidates(snapshots, Range::MIN);
 	if (candidates.empty())
 	{
-	    // not enough candidates to satisfy quota
+	    // not enough candidates to satisfy the condition
+
+#ifdef VERBOSE_LOGGING
+	    cout << "condition not satisfied" << '\n';
+#endif
+
 	    return;
 	}
 
@@ -293,17 +357,27 @@ Cleaner::cleanup_quota_aware(ProxySnapshots& snapshots)
 	    if (!tmp.empty())
 	    {
 		remove(tmp);
-		// after removing snapshots is_quota_satisfied must be reevaluated
+
+		// after removing snapshots the condition must be reevaluated
 		break;
 	    }
 
 	    if (next(e) == candidates.end())
 	    {
-		// not enough candidates to satisfy quota
+		// not enough candidates to satisfy the condition
+
+#ifdef VERBOSE_LOGGING
+		cout << "condition not satisfied" << '\n';
+#endif
+
 		return;
 	    }
 	}
     }
+
+#ifdef VERBOSE_LOGGING
+    cout << "condition satisfied" << '\n';
+#endif
 }
 
 
@@ -312,10 +386,54 @@ Cleaner::cleanup()
 {
     ProxySnapshots& snapshots = snapper->getSnapshots();
 
-    cleanup_quota_unaware(snapshots);
+#ifdef VERBOSE_LOGGING
+    cout << "cleanup without condition" << '\n';
+#endif
+
+    cleanup(snapshots);
 
     if (is_quota_aware())
-	cleanup_quota_aware(snapshots);
+    {
+#ifdef VERBOSE_LOGGING
+	cout << "cleanup with quota condition" << '\n';
+#endif
+
+	cleanup(snapshots, std::bind(&Cleaner::is_quota_satisfied, this));
+    }
+    else
+    {
+#ifdef VERBOSE_LOGGING
+	cout << "no cleanup with quota condition" << '\n';
+#endif
+    }
+
+    if (is_free_aware())
+    {
+#ifdef VERBOSE_LOGGING
+	cout << "cleanup with free condition" << '\n';
+#endif
+
+	cleanup(snapshots, std::bind(&Cleaner::is_free_satisfied, this));
+    }
+    else
+    {
+#ifdef VERBOSE_LOGGING
+	cout << "no cleanup with free condition" << '\n';
+#endif
+    }
+}
+
+
+void
+Cleaner::cleanup(std::function<bool()> condition)
+{
+    ProxySnapshots& snapshots = snapper->getSnapshots();
+
+#ifdef VERBOSE_LOGGING
+    cout << "cleanup with user condition" << '\n';
+#endif
+
+    cleanup(snapshots, condition);
 }
 
 
@@ -323,7 +441,7 @@ struct NumberParameters : public Parameters
 {
     NumberParameters(const ProxySnapper* snapper);
 
-    bool is_degenerated() const;
+    bool is_degenerated() const override;
 
     Range limit;
     Range limit_important;
@@ -345,8 +463,8 @@ NumberParameters::NumberParameters(const ProxySnapper* snapper)
 ostream&
 operator<<(ostream& s, const NumberParameters& parameters)
 {
-    return s << dynamic_cast<const Parameters&>(parameters) << endl
-	     << "limit:" << parameters.limit << endl
+    return s << dynamic_cast<const Parameters&>(parameters) << '\n'
+	     << "limit:" << parameters.limit << '\n'
 	     << "limit-important:" << parameters.limit_important;
 }
 
@@ -377,7 +495,7 @@ private:
 
 
     list<ProxySnapshots::iterator>
-    calculate_candidates(ProxySnapshots& snapshots, Range::Value value) override
+    calculate_candidates(ProxySnapshots& snapshots, const Range::Value& value) override
     {
 	const NumberParameters& parameters = dynamic_cast<const NumberParameters&>(Cleaner::parameters);
 
@@ -430,11 +548,20 @@ do_cleanup_number(ProxySnapper* snapper, bool verbose)
 }
 
 
+void
+do_cleanup_number(ProxySnapper* snapper, bool verbose, std::function<bool()> condition)
+{
+    NumberParameters parameters(snapper);
+    NumberCleaner cleaner(snapper, verbose, parameters);
+    cleaner.cleanup(condition);
+}
+
+
 struct TimelineParameters : public Parameters
 {
     TimelineParameters(const ProxySnapper* snapper);
 
-    bool is_degenerated() const;
+    bool is_degenerated() const override;
 
     Range limit_hourly;
     Range limit_daily;
@@ -463,11 +590,11 @@ TimelineParameters::TimelineParameters(const ProxySnapper* snapper)
 ostream&
 operator<<(ostream& s, const TimelineParameters& parameters)
 {
-    return s << dynamic_cast<const Parameters&>(parameters) << endl
-	     << "limit-hourly:" << parameters.limit_hourly << endl
-	     << "limit-daily:" << parameters.limit_daily << endl
-	     << "limit-weekly:" << parameters.limit_weekly << endl
-	     << "limit-monthly:" << parameters.limit_monthly << endl
+    return s << dynamic_cast<const Parameters&>(parameters) << '\n'
+	     << "limit-hourly:" << parameters.limit_hourly << '\n'
+	     << "limit-daily:" << parameters.limit_daily << '\n'
+	     << "limit-weekly:" << parameters.limit_weekly << '\n'
+	     << "limit-monthly:" << parameters.limit_monthly << '\n'
 	     << "limit-yearly:" << parameters.limit_yearly;
 }
 
@@ -562,7 +689,7 @@ private:
 
 
     list<ProxySnapshots::iterator>
-    calculate_candidates(ProxySnapshots& snapshots, Range::Value value) override
+    calculate_candidates(ProxySnapshots& snapshots, const Range::Value& value) override
     {
 	const TimelineParameters& parameters = dynamic_cast<const TimelineParameters&>(Cleaner::parameters);
 
@@ -633,6 +760,15 @@ do_cleanup_timeline(ProxySnapper* snapper, bool verbose)
 }
 
 
+void
+do_cleanup_timeline(ProxySnapper* snapper, bool verbose, std::function<bool()> condition)
+{
+    TimelineParameters parameters(snapper);
+    TimelineCleaner cleaner(snapper, verbose, parameters);
+    cleaner.cleanup(condition);
+}
+
+
 struct EmptyPrePostParameters : public Parameters
 {
     EmptyPrePostParameters(const ProxySnapper* snapper);
@@ -659,7 +795,7 @@ public:
 private:
 
     list<ProxySnapshots::iterator>
-    calculate_candidates(ProxySnapshots& snapshots, Range::Value value) override
+    calculate_candidates(ProxySnapshots& snapshots, const Range::Value& value) override
     {
 	list<ProxySnapshots::iterator> ret;
 
@@ -692,3 +828,12 @@ do_cleanup_empty_pre_post(ProxySnapper* snapper, bool verbose)
     EmptyPrePostCleaner cleaner(snapper, verbose, parameters);
     cleaner.cleanup();
 }
+
+
+void
+do_cleanup_empty_pre_post(ProxySnapper* snapper, bool verbose, std::function<bool()> condition)
+{
+    EmptyPrePostParameters parameters(snapper);
+    EmptyPrePostCleaner cleaner(snapper, verbose, parameters);
+    cleaner.cleanup(condition);
+}
