@@ -1,5 +1,6 @@
 /*
  * Copyright (c) [2011-2014] Novell, Inc.
+ * Copyright (c) 2018 SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -27,6 +28,19 @@ using std::string;
 #include "dbus/DBusConnection.h"
 
 
+/**
+ * Translate an DBus exception to the corresponding snapper exception
+ * (iff such exists). Unfinished.
+ */
+void
+convert_exception(const DBus::ErrorException& e) __attribute__ ((__noreturn__));
+
+
+/**
+ * Returns a string explaining the DBus exception. Function should
+ * likely be removed once convert_exception is complete and used
+ * everywhere.
+ */
 string
 error_description(const DBus::ErrorException& e);
 
