@@ -147,10 +147,14 @@ namespace snapper
     string
     prepend_root_prefix(const string& root_prefix, const string& path)
     {
-        if (root_prefix == "/")
-            return path;
-        else
-            return root_prefix + path;
+	// TODO use std::filesystem (C++17)
+
+	if (root_prefix == "/" || root_prefix.empty())
+	    return path;
+	else if (path == "/")
+	    return root_prefix;
+	else
+	    return root_prefix + path;
     }
 
 
@@ -220,22 +224,6 @@ namespace snapper
     }
 
 
-    string
-    stringerror(int errnum)
-    {
-#if (_POSIX_C_SOURCE >= 200112L) && ! _GNU_SOURCE
-	char buf1[100];
-	if (strerror_r(errnum, buf1, sizeof(buf1) - 1) == 0)
-	    return string(buf1);
-	return string("strerror failed");
-#else
-	char buf1[100];
-	const char* buf2 = strerror_r(errnum, buf1, sizeof(buf1) - 1);
-	return string(buf2);
-#endif
-    }
-
-
     string
     sformat(const char* format, ...)
     {
@@ -285,10 +273,10 @@ namespace snapper
     {
 	struct tm s;
 	memset(&s, 0, sizeof(s));
-	const char* p = strptime(str.c_str(), "%F %T", &s);
+	const char* p = strptime(str.c_str(), "%Y-%m-%d %T", &s);
 	if (!p || *p != '\0')
 	    return (time_t)(-1);
-	return utc ? timegm(&s) : timelocal(&s);
+	return utc ? timegm(&s) : mktime(&s);
     }
 
 
@@ -315,8 +303,6 @@ namespace snapper
 	if (e != 0 || result == NULL)
 	    return false;
 
-	memset(pwd.pw_passwd, 0, strlen(pwd.pw_passwd));
-
 	username = pwd.pw_name;
 	gid = pwd.pw_gid;
 
@@ -324,6 +310,27 @@ namespace snapper
     }
 
 
+    bool
+    get_uid_dir(uid_t uid, string& dir)
+    {
+	struct passwd pwd;
+	struct passwd* result;
+
+	vector<char> buf(sysconf(_SC_GETPW_R_SIZE_MAX, 1024));
+
+	int e;
+	while ((e = getpwuid_r(uid, &pwd, buf.data(), buf.size(), &result)) == ERANGE)
+	    buf.resize(2 * buf.size());
+
+	if (e != 0 || result == NULL)
+	    return false;
+
+	dir = pwd.pw_dir;
+
+	return true;
+    }
+
+
     bool
     get_user_uid(const char* username, uid_t& uid)
     {
@@ -342,8 +349,6 @@ namespace snapper
 	    return false;
 	}
 
-	memset(pwd.pw_passwd, 0, strlen(pwd.pw_passwd));
-
 	uid = pwd.pw_uid;
 
 	return true;
@@ -368,8 +373,6 @@ namespace snapper
 	    return false;
 	}
 
-	memset(grp.gr_passwd, 0, strlen(grp.gr_passwd));
-
 	gid = grp.gr_gid;
 
 	return true;
